import os.path as os_path
import pkg_resources
import traceback

import nanite
from PyQt5 import uic, QtCore, QtWidgets
from PyQt5.QtCore import QStandardPaths

from ..extensions import ExtensionManager, SUPPORTED_FORMATS


class ExtensionErrorWrapper:
    def __init__(self, ehash):
        self.ehash = ehash

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, trc):
        if exc_type is not None:
            QtWidgets.QMessageBox.warning(
                None,
                f"Loading extension {self.ehash} failed!",
                f"It was not possible to load the extension {self.ehash}! "
                + "You might have to install additional software:\n\n"
                + traceback.format_exc(),
                )
            return True  # do not raise the exception


class Preferences(QtWidgets.QDialog):
    """Implements the plotting pipeline using pyqtgraph"""
    instances = {}
    feature_changed = QtCore.pyqtSignal()

    def __init__(self, parent, *args, **kwargs):
        QtWidgets.QWidget.__init__(self, parent=parent, *args, **kwargs)
        path_ui = pkg_resources.resource_filename(
            "pyjibe.head", "preferences.ui")
        uic.loadUi(path_ui, self)
        self.settings = QtCore.QSettings()
        self.parent = parent

        #: configuration keys, corresponding widgets, and defaults
        self.config_pairs = [
            ["advanced/developer mode", self.advanced_developer_mode, 0],
            ["advanced/expert mode", self.advanced_expert_mode, 0],
            ["check for updates", self.general_check_for_updates, 1],
        ]

        # extensions
        store_path = os_path.join(
            QStandardPaths.writableLocation(
                QStandardPaths.AppDataLocation), "extensions")
        self.extensions = ExtensionManager(store_path)

        self.reload()

        # signals
        self.btn_apply = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Apply)
        self.btn_apply.clicked.connect(self.on_settings_apply)
        self.btn_cancel = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.Cancel)
        self.btn_ok = self.buttonBox.button(QtWidgets.QDialogButtonBox.Ok)
        self.btn_ok.clicked.connect(self.on_settings_apply)
        self.btn_restore = self.buttonBox.button(
            QtWidgets.QDialogButtonBox.RestoreDefaults)
        self.btn_restore.clicked.connect(self.on_settings_restore)
        # extension buttons
        self.checkBox_ext_enabled.clicked.connect(self.on_ext_enabled)
        self.pushButton_ext_load.clicked.connect(self.on_ext_load)
        self.pushButton_ext_remove.clicked.connect(self.on_ext_remove)
        self.listWidget_ext.currentRowChanged.connect(self.on_ext_selected)
        self.listWidget_ext.itemChanged.connect(self.on_ext_modified)
        # tab changed
        self.tabWidget.currentChanged.connect(self.on_tab_changed)

    def reload(self):
        """Read configuration or set default parameters"""
        for key, widget, default in self.config_pairs:
            value = self.settings.value(key, default)
            if isinstance(widget, QtWidgets.QCheckBox):
                widget.setChecked(bool(int(value)))
            elif isinstance(widget, QtWidgets.QLineEdit):
                widget.setText(value)
            elif widget is self.dcor_servers:
                self.dcor_servers.clear()
                self.dcor_servers.addItems(value)
                self.dcor_servers.setCurrentIndex(0)
            else:
                raise NotImplementedError("No rule for '{}'".format(key))

        # peculiarities of developer mode
        devmode = bool(int(self.settings.value("advanced/developer mode", 0)))
        if devmode:
            nanite.read.DEFAULT_MODALITY = None
        else:
            nanite.read.DEFAULT_MODALITY = "force-distance"

        self.reload_ext()

    def reload_ext(self):
        """Reload the list of extensions"""
        # extensions
        row = self.listWidget_ext.currentRow()
        self.listWidget_ext.blockSignals(True)
        self.listWidget_ext.clear()
        have_extensions = bool(self.extensions)
        self.widget_ext_controls.setVisible(have_extensions)
        if have_extensions:
            for ii, ext in enumerate(self.extensions):
                lwitem = QtWidgets.QListWidgetItem(ext.title,
                                                   self.listWidget_ext)
                lwitem.setFlags(QtCore.Qt.ItemIsEditable
                                | QtCore.Qt.ItemIsSelectable
                                | QtCore.Qt.ItemIsEnabled
                                | QtCore.Qt.ItemIsUserCheckable)
                lwitem.setCheckState(2 if ext.enabled else 0)
                lwitem.setData(100, ext.hash)
            self.listWidget_ext.setCurrentRow(0)
            if row + 1 > self.listWidget_ext.count() or row < 0:
                row = 0
            self.listWidget_ext.setCurrentRow(row)
        self.listWidget_ext.blockSignals(False)
        self.on_ext_selected()

    @QtCore.pyqtSlot(bool)
    def on_ext_enabled(self, enabled):
        """Enable or disable an extension (signal from checkbox widget)"""
        item = self.listWidget_ext.currentItem()
        ehash = item.data(100)
        with ExtensionErrorWrapper(ehash):
            self.extensions.extension_set_enabled(ehash, enabled)
        self.reload_ext()
        self.feature_changed.emit()

    @QtCore.pyqtSlot()
    def on_ext_load(self):
        """Load an extension from the file system"""
        format_string = " ".join(f"*{su}" for su in SUPPORTED_FORMATS)
        paths, _ = QtWidgets.QFileDialog.getOpenFileNames(
            parent=self,
            caption="Select an extension file",
            directory=self.settings.value("paths/extension", ""),
            filter=f"Supported extension files ({format_string})")
        if paths:
            for pp in paths:
                with ExtensionErrorWrapper(pp):
                    self.extensions.import_extension_from_path(pp)
            self.reload_ext()
            self.feature_changed.emit()

    @QtCore.pyqtSlot()
    def on_ext_remove(self):
        """Remove an extension"""
        ehash = self.listWidget_ext.currentItem().data(100)
        self.extensions.extension_remove(ehash)
        self.reload_ext()
        self.feature_changed.emit()

    @QtCore.pyqtSlot(QtWidgets.QListWidgetItem)
    def on_ext_modified(self, item):
        """Enable or disable an extension (signal from listWidget)"""
        ehash = item.data(100)
        enabled = bool(item.checkState())
        with ExtensionErrorWrapper(ehash):
            self.extensions.extension_set_enabled(ehash, enabled)
        self.listWidget_ext.setCurrentItem(item)
        self.reload_ext()
        self.feature_changed.emit()

    @QtCore.pyqtSlot()
    def on_ext_selected(self):
        """Display details for an extension (signal from listWidget)"""
        item = self.listWidget_ext.currentItem()
        if item is not None:
            ehash = item.data(100)
            ext = self.extensions[ehash]
            self.checkBox_ext_enabled.blockSignals(True)
            self.checkBox_ext_enabled.setChecked(ext.enabled)
            self.checkBox_ext_enabled.blockSignals(False)
            with ExtensionErrorWrapper(ehash):
                if ext.enabled:  # only load the extension if enabled
                    ext.load()
            self.label_ext_name.setText(ext.title)
            item.setText(ext.title)
            self.label_ext_description.setText(ext.description)

    @QtCore.pyqtSlot()
    def on_settings_apply(self):
        """Save current changes made in UI to settings and reload UI"""
        for key, widget, default in self.config_pairs:
            if isinstance(widget, QtWidgets.QCheckBox):
                value = int(widget.isChecked())
                if key == "advanced/developer mode":
                    devmode = int(self.settings.value(
                        "advanced/developer mode", 0))
                    if devmode != value:
                        msg = QtWidgets.QMessageBox()
                        msg.setIcon(QtWidgets.QMessageBox.Information)
                        msg.setText("Please restart PyJibe for the changes "
                                    + "to take effect.")
                        msg.setWindowTitle("Restart PyJibe")
                        msg.exec_()
            elif isinstance(widget, QtWidgets.QLineEdit):
                value = widget.text().strip()
            elif widget is self.dcor_servers:
                curtext = self.dcor_servers.currentText()
                items = self.settings.value(key, default)
                if curtext in items:
                    items.remove(curtext)
                items.insert(0, curtext)
                value = items
            else:
                raise NotImplementedError("No rule for '{}'".format(key))
            self.settings.setValue(key, value)

        # reload UI to give visual feedback
        self.reload()

    @QtCore.pyqtSlot()
    def on_settings_restore(self):
        self.settings.clear()
        self.reload()

    @QtCore.pyqtSlot()
    def on_tab_changed(self):
        if self.tabWidget.currentWidget() is self.tab_extensions:
            # Managing extensions has nothing to do with other settings.
            enabled = False
        else:
            enabled = True

        self.btn_apply.setEnabled(enabled)
        self.btn_cancel.setEnabled(enabled)
        self.btn_restore.setEnabled(enabled)



"""Test of data set functionalities"""
import numpy as np
import pytest

import lmfit
from nanite.model import model_hertz_paraboloidal

import nanite.model.residuals as nres
import pyjibe.head

from helpers import MockModelModule, make_directory_with_data


def test_simple(qtbot):
    """Open the main window and close it again"""
    main_window = pyjibe.head.PyJibe()
    main_window.close()


def test_clear_and_verify_data(qtbot):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data())
    war = main_window.subwindows[0].widget()
    # clear data
    tpp = war.tab_preprocess
    war.cb_autosave.setChecked(0)
    # perform simple filter
    tpp.set_preprocessing(["compute_tip_position"])
    # perform fitting with standard parameters
    # set initial parameters in user interface
    itab = war.tab_fit.table_parameters_initial
    # disable weighting
    war.tab_fit.cb_weight_cp.setCheckState(0)
    # enable fitting of force offset
    itab.item(4, 0).setCheckState(0)
    # set better value for contact point
    itab.item(3, 1).setText(str(18000))
    apret = war.data_set[0]
    war.tab_fit.fit_approach_retract(apret)
    # Now check something
    ftab = war.tab_fit.table_parameters_fitted
    # E
    assert np.allclose(apret.fit_properties["params_fitted"]["E"].value,
                       14741.256863072116,
                       atol=1,
                       rtol=0)
    assert float(ftab.item(0, 0).text()) == 14741
    # contact_point
    assert float(ftab.item(1, 0).text()) == 18029
    # baseline_offset
    assert float(ftab.item(2, 0).text()) == -480.67


def test_fit_all(qtbot):
    """Perform a simple fit with the standard parameters"""
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data(2))
    war = main_window.subwindows[0].widget()
    war.cb_autosave.setChecked(0)
    war.on_fit_all()
    a1 = war.data_set[0]
    a2 = war.data_set[1]
    assert a1.fit_properties == a2.fit_properties


def test_hidden_parameters(qtbot):
    def get_parameter_defaults():
        params = lmfit.Parameters()
        params.add("E", value=3e3, min=0)
        params.add("R", value=10e-6, min=0, vary=False)
        params.add("_nu", value=.5, min=0, max=0.5, vary=False)
        params.add("contact_point", value=0)
        params.add("baseline", value=0)
        return params

    def model_func(delta, E, R, _nu, contact_point=0, baseline=0):
        return model_hertz_paraboloidal.hertz_paraboloidal(
            delta=delta,
            E=E,
            R=R,
            nu=_nu,
            contact_point=contact_point,
            baseline=baseline)
    with MockModelModule(
            model_key="petperpan",
            parameter_keys=["E", "R", "_nu", "contact_point", "baseline"],
            model_func=model_func,
            model=nres.get_default_modeling_wrapper(model_func),
            residual=nres.get_default_residuals_wrapper(model_func),
            get_parameter_defaults=get_parameter_defaults) as mod:
        main_window = pyjibe.head.PyJibe()
        # disable developer mode
        main_window.settings.setValue("advanced/developer mode", 0)
        main_window.load_data(files=make_directory_with_data(2))
        war = main_window.subwindows[0].widget()
        # clear data
        war.cb_autosave.setChecked(0)
        # perform simple filter
        war.tab_preprocess.set_preprocessing(["compute_tip_position"])
        # set mock model
        idx = war.tab_fit.cb_model.findData(mod.model_key)
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # Check visibility of Poisson's ratio
        itab = war.tab_fit.table_parameters_initial
        assert itab.rowCount() == 4


def test_hidden_parameters_control_in_dev_mode(qtbot):
    def get_parameter_defaults():
        params = lmfit.Parameters()
        params.add("E", value=3e3, min=0)
        params.add("R", value=10e-6, min=0, vary=False)
        params.add("_nu", value=.5, min=0, max=0.5, vary=False)
        params.add("contact_point", value=0)
        params.add("baseline", value=0)
        return params

    def model_func(delta, E, R, _nu, contact_point=0, baseline=0):
        return model_hertz_paraboloidal.hertz_paraboloidal(
            delta=delta,
            E=E,
            R=R,
            nu=_nu,
            contact_point=contact_point,
            baseline=baseline)
    with MockModelModule(
            model_key="petperpan",
            parameter_keys=["E", "R", "_nu", "contact_point", "baseline"],
            model_func=model_func,
            model=nres.get_default_modeling_wrapper(model_func),
            residual=nres.get_default_residuals_wrapper(model_func),
            get_parameter_defaults=get_parameter_defaults) as mod:
        main_window = pyjibe.head.PyJibe()
        # disable developer mode
        main_window.settings.setValue("advanced/developer mode", 1)
        main_window.load_data(files=make_directory_with_data(2))
        war = main_window.subwindows[0].widget()
        # clear data
        war.cb_autosave.setChecked(0)
        # perform simple filter
        war.tab_preprocess.set_preprocessing(["compute_tip_position"])
        # set mock model
        idx = war.tab_fit.cb_model.findData(mod.model_key)
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # Check visibility of Poisson's ratio
        itab = war.tab_fit.table_parameters_initial
        assert itab.rowCount() == 5


@pytest.mark.parametrize("method,contact_point", [
    ["gradient_zero_crossing", 1895],
    ["fit_constant_line", 1919],
    ["deviation_from_baseline", 1908],
    ])
def test_preprocessing_poc_estimation(method, contact_point):
    main_window = pyjibe.head.PyJibe()
    main_window.load_data(files=make_directory_with_data(2))
    war = main_window.subwindows[0].widget()
    # clear data
    war.cb_autosave.setChecked(0)
    war.tabs.setCurrentIndex(0)
    # perform simple filter
    war.tab_preprocess.set_preprocessing(
        preprocessing=["compute_tip_position", "correct_tip_offset"],
        options={"correct_tip_offset": {"method": method}}
    )
    war.tabs.setCurrentIndex(1)
    fd = war.data_set[0]
    assert np.argmin(np.abs(fd["tip position"])) == contact_point



import collections

from cycler import cycler
import matplotlib.pylab as plt
from matplotlib.figure import Figure
from matplotlib.backends.backend_qt5agg import (
    FigureCanvasQTAgg as FigureCanvas)
import numpy as np

from ..head import custom_widgets


class MPLPreproc:
    def __init__(self):
        """Matplotlib plot for preprocessing data"""
        # Add matplotlib figure
        self.figure = Figure(facecolor="none", tight_layout=True,
                             frameon=True)
        self.axis = self.figure.add_subplot(111)
        self.twin_axes = collections.OrderedDict()
        self.axis.set_facecolor('#FFFFFF')
        self.canvas = FigureCanvas(self.figure)
        self.canvas.draw()

    def add_toolbar(self, widget):
        """Add toolbar to PyQT widget"""
        self.toolbar = custom_widgets.NavigationToolbarPreproc(
            self.canvas,
            widget,
            coordinates=True
        )

    def get_normed_axis(self, name):
        """Get the proper axis to plot to

        All axes share the same x-axis, but might have a different y-axis.
        We identify the differently normed axis via `name`.
        """
        if name not in self.twin_axes:
            if not self.twin_axes:
                # No twin axis has been created yet, so populate the
                # dictionary with the initial axis.
                self.twin_axes[name] = self.axis
            else:
                self.twin_axes[name] = self.axis.twinx()
        # Let all plots share the same color cycler (so we don't have
        # the same color for different lines).
        # Get the number of lines currently plotted.
        nl = self.get_num_lines()
        colors = plt.rcParams["axes.prop_cycle"].by_key()["color"]
        new_colors = np.roll(colors, -nl).tolist()
        self.twin_axes[name].set_prop_cycle(
            cycler(color=new_colors))
        return self.twin_axes[name]

    def get_num_lines(self):
        lines = []
        for ax in self.twin_axes:
            lines2, _ = self.twin_axes[ax].get_legend_handles_labels()
            lines += lines2
        return len(lines)

    def clear(self):
        self.axis.clear()
        for ax in self.twin_axes:
            self.twin_axes[ax].clear()
        self.axis.grid()

    def legend(self):
        lines, labels = [], []
        for ax in self.twin_axes:
            lines2, labels2 = self.twin_axes[ax].get_legend_handles_labels()
            lines += lines2
            labels += labels2
        if self.twin_axes:
            axtop = list(self.twin_axes.values())[-1]
            axtop.legend(lines, labels)



"""Test model with expression parameters"""
import numpy as np

import pyjibe.head

from helpers import MockModelModuleExpr, make_directory_with_data


def test_model_expr_base(qtbot):
    """Expressions sanity check

    This test is just to make sure that parameters that have an
    expression are not overridden by the UI.
    """
    with MockModelModuleExpr() as mod:
        main_window = pyjibe.head.PyJibe()
        main_window.load_data(files=make_directory_with_data(2))
        war = main_window.subwindows[0].widget()
        # clear data
        war.cb_autosave.setChecked(0)
        # perform simple filter
        war.tab_preprocess.set_preprocessing(["compute_tip_position"])
        # set mock model
        idx = war.tab_fit.cb_model.findData(mod.model_key)
        war.tab_fit.cb_model.setCurrentIndex(idx)
        # Now there should be fit results
        itab = war.tab_fit.table_parameters_initial
        # set better value for contact point
        itab.item(5, 1).setText(str(18000))
        fdist = war.data_set[0]
        parmsi = fdist.fit_properties["params_initial"]
        parmsf = fdist.fit_properties["params_fitted"]
        assert parmsi["E1"].expr == "virtual_parameter+E"
        assert parmsf["E1"].expr == "virtual_parameter+E"
        assert np.allclose(parmsf["E1"].value, 15388.787369767488,
                           atol=10)  # this fit is not very stable



import collections
import functools
import pathlib
import shutil

from nanite import model as nmodel

from .util import hashfile


SUPPORTED_FORMATS = [
    ".py",
]


class ExtensionManager:
    def __init__(self, store_path):
        """Extension manager

        This class can be used to maintain a set of extensions
        (e.g. fitting models) at a specific location `store_path`
        in the file system.

        All extensions are loaded during instantiation. It is
        possible to disable individual extensions (without
        deleting them) on-the-fly.
        """
        self.store_path = pathlib.Path(store_path)
        self.store_path.mkdir(exist_ok=True, parents=True)
        self.extension_hash_dict = collections.OrderedDict()
        self.load_extensions_from_store()

    def __contains__(self, key):
        if isinstance(key, Extension):
            key = key.hash
        return key in self.extension_hash_dict

    def __getitem__(self, key):
        return self.get_extension_or_bust(key)

    def __iter__(self):
        for ahash in self.extension_hash_dict:
            yield self.extension_hash_dict[ahash]

    def __len__(self):
        return len(self.extension_hash_dict)

    def get_extension_or_bust(self, ext):
        """Return an Extension instance or raise ValueError

        Parameter `ext` can be an instance of Extension, an
        extension hash, or an index. If an Extension or a
        hash is provided, this function checks whether the
        extension is maintained by this manager and returns
        the correct instance.

        Parameters
        ----------
        ext: Extension
            Extension instance
        """
        if isinstance(ext, Extension):
            the_hash = ext.hash
        elif isinstance(ext, int):
            the_hash = list(self.extension_hash_dict.keys())[ext]
        else:
            the_hash = ext
        if the_hash in self.extension_hash_dict:
            return self.extension_hash_dict[the_hash]
        else:
            raise ValueError(f"Extension not in {self.store_path}: {the_hash}")

    def import_extension_from_path(self, path):
        """Import an extension file to `self.store_path`"""
        ext = Extension(path)
        if ext.hash not in self.extension_hash_dict:
            new_name = f"ext_{ext.type}_{ext.hash}{ext.suffix}"
            new_path = self.store_path / new_name
            shutil.copy2(path, new_path)
            ext.path = new_path
            self.extension_load(ext)
        return ext

    def load_extensions_from_store(self):
        """Load all extensions from `self.store_path`

        This function is called during initialization.
        """
        failed = []
        # load all (enabled) extensions
        for pp in self.store_path.glob("ext_*"):
            if pp.suffix in SUPPORTED_FORMATS:
                ext = Extension(pp)
                try:
                    self.extension_load(ext)
                except BaseException as e:
                    failed.append([ext, f"{e.__class__.__name__}: {e}"])
        if failed:
            raise ValueError(f"Could not load these extensions: {failed}")

    def extension_load(self, ext):
        """Load a single extension

        Parameters
        ----------
        ext: Extension
            Extension instance
        """
        # add to hash dict first in case it gets disabled
        self.extension_hash_dict[ext.hash] = ext
        ext.load()

    def extension_remove(self, ext):
        """Remove an extension from `self.store_path` and unload it

        Parameters
        ----------
        ext: Extension or str or int
            Extension instance, its hash or its index in
            ExtensionManager
        """
        ext = self.get_extension_or_bust(ext)
        self.extension_hash_dict.pop(ext.hash)
        # re-instantiate ext to get the path right
        ext.destroy()

    def extension_set_enabled(self, ext, enabled):
        """Enable or disable an extension

        Parameters
        ----------
        ext: Extension or str or int
            Extension instance, its hash or its index in
            ExtensionManager
        enabled: bool
            Whether to enable the extension.
        """
        ext = self.get_extension_or_bust(ext)
        ext.set_enabled(enabled)


class Extension:
    def __init__(self, path):
        """Helper class for managing individual extensions"""
        self.path = pathlib.Path(path)
        self.suffix = self.path.suffix
        self.data = {}

    def __repr__(self):
        return f"<PyJibe Extension {self.path} at {hex(id(self))}>"

    @property
    def description(self):
        """Description of the extension"""
        description = "No description provided."
        if self.loaded:
            description = self.data["model"].model_doc
        return description

    @property
    def enabled(self):
        """Whether the extension is enabled"""
        return not self.path_lock_disabled.exists()

    @property
    @functools.lru_cache()
    def hash(self):
        """MD5 hash of the extension"""
        return hashfile(self.path)

    @property
    def loaded(self):
        """Whether the extension is currently loaded"""
        return "model" in self.data

    @property
    def path_lock_disabled(self):
        return self.path.with_name(self.path.name + "_disabled")

    @property
    def title(self):
        """Descriptive title including version of the extension"""
        title = self.path.name  # fallback
        if self.loaded:
            title = self.data["model"].model_name
        return title

    @property
    @functools.lru_cache()
    def type(self):
        """Type of the extension (e.g. "feat_anc_plugin")"""
        if self.path.suffix == ".py":
            return "nanite_model"
        else:
            raise ValueError(f"Cannot determine extension type: {self.path}!")

    def set_enabled(self, enabled):
        """Set this extension to enabled (True) or disabled (False)

        The extension is also loaded (True) or unloaded (False).
        """
        if enabled:
            # set enabled so that `load` works
            self.path_lock_disabled.unlink(missing_ok=True)
            self.load()
        else:
            self.unload()
            self.path_lock_disabled.touch()

    def load(self):
        """Load the extension if it is enabled and not loaded"""
        if not self.enabled or self.loaded:
            # do not load disabled extensions or extensions already loaded
            return
        try:
            # import the file
            mod = nmodel.load_model_from_file(self.path, register=True)
            self.data["model"] = mod
        except BaseException:
            # If loading the extension fails, disable it and only then
            # raise the exception.
            self.set_enabled(False)
            raise

    def unload(self):
        """Unload the extension"""
        if "model" in self.data:
            nmodel.deregister_model(self.data["model"])

    def destroy(self):
        """Unload and remove the extension"""
        self.unload()
        self.path_lock_disabled.unlink(missing_ok=True)
        self.path.unlink(missing_ok=True)



import functools
import hashlib
import pathlib
import warnings

import numpy as np


def hashfile(fname, blocksize=65536, count=0, constructor=hashlib.md5,
             hasher_class=None):
    """Compute md5 hex-hash of a file

    Parameters
    ----------
    fname: str or pathlib.Path
        path to the file
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    hasher_class: callable
        deprecated, see use `constructor` instead
    constructor: callable
        hash algorithm constructor
    """
    if hasher_class is not None:
        warnings.warn("The `hasher_class` argument is deprecated, please use "
                      "`constructor` instead.")
        constructor = hasher_class
    path = pathlib.Path(fname).resolve()
    return _hashfile_cached(
        path=path,
        stat=path.stat(),
        blocksize=blocksize,
        count=count,
        constructor=constructor
    )


@functools.lru_cache(maxsize=100)
def _hashfile_cached(path, stat, blocksize=65536, count=0,
                     constructor=hashlib.md5):
    """Cached hashfile using stat tuple as cache

    This is a privat function. Please use `hashfile` instead!

    Parameters
    ----------
    path: pathlib.Path
        path to the file to be hashed
    stat: named tuple
        tuple of `os.stat_result` for `path`. This must be specified,
        so that caching of the result is done properly in case the user
        modified `path` (this function is wrapped with
        functools.lru_cache)
    blocksize: int
        block size in bytes read from the file
        (set to `0` to hash the entire file)
    count: int
        number of blocks read from the file
    constructor: callable
        hash algorithm constructor
    """
    assert stat, "We need stat for validating the cache"
    hasher = constructor()
    with path.open('rb') as fd:
        buf = fd.read(blocksize)
        ii = 0
        while len(buf) > 0:
            hasher.update(buf)
            buf = fd.read(blocksize)
            ii += 1
            if count and ii == count:
                break
    return hasher.hexdigest()


def equals(a, b):
    """Check if `a` equals `b` with support for arrays"""
    if a.__class__ != b.__class__:
        return False
    if isinstance(a, dict):
        if len(a) != len(b):
            return False
        ret = True
        for key in a:
            if key not in b:
                return False
            else:
                ret *= equals(a[key], b[key])
        return ret
    elif isinstance(a, (list, tuple)):
        if len(a) != len(b):
            return False
        ret = True
        for ii in range(len(a)):
            ret *= equals(a[ii], b[ii])
        return ret
    elif isinstance(a, np.ndarray):
        if len(a) != len(b):
            return False
        else:
            return np.all(a == b)
    else:
        return a == b
